__all__ = ('DatagramAddressedReadProtocol', 'DatagramMergerReadProtocol', 'ReadProtocolBase', 'ReadWriteProtocolBase',)

from collections import deque as Deque
from functools import partial as partial_func
from warnings import warn

from ...utils import copy_docs, to_coroutine

from ..traps import Future, Task, skip_ready_cycle

from .abstract import AbstractProtocolBase
from .payload_stream import PayloadStream


BUFFER_SIZE_MAX = 131072


def _get_end_intersection_sizes(chunk, offset, boundary):
    """
    Gets the end intersection sizes.
    
    Parameters
    ----------
    chunk : `bytes`
        Data chunk to process.
    
    offset : `int`
        Data chunk offset.
    
    boundary : `bytes`
        Boundary to check intersection with.
    
    Returns
    -------
    intersection_sizes : `None | list<int>`
    """
    intersection_sizes = None
    
    chunk_length = len(chunk)
    boundary_length = len(boundary) - 1
    
    chunk_start_index = chunk_length - boundary_length
    if chunk_start_index < offset:
        chunk_start_index = offset
    elif chunk_start_index < 0:
        chunk_start_index = 0
    
    for chunk_start_index in range(chunk_start_index, chunk_length):
        for shift in range(0, chunk_length - chunk_start_index):
            if chunk[chunk_start_index + shift] != boundary[shift]:
                break
        
        else:
            if intersection_sizes is None:
                intersection_sizes = []
            
            intersection_sizes.append(chunk_length - chunk_start_index)
    
    return intersection_sizes


def _finish_intersection_sizes(chunk, boundary, intersection_sizes):
    """
    Finishes intersection sizes.
    
    Parameters
    ----------
    chunk : `bytes`
        Data chunk to process.
    
    boundary : `bytes`
        Boundary to check intersection with.
    
    intersection_sizes : `list<int>`
        Detected intersection sizes of on the previous chunk.
        The processed ones are removed from it while the too low values to finish processing on are left in it.
    
    Returns
    -------
    consumed_bytes : `int`
        Returns `-1` on failure.
    intersection_sizes : `None | list<int>`
        Leftover intersection sizes if chunk size is too small.
    """
    chunk_length = len(chunk)
    boundary_length = len(boundary) - 1
    continued_intersection_sizes = None
    
    for intersection_size in intersection_sizes:
        leftover_size = boundary_length - intersection_size + 1
        if leftover_size > chunk_length:
            if continued_intersection_sizes is None:
                continued_intersection_sizes = []
            continued_intersection_sizes.append(intersection_size)
            continue
        
        for shift in range(0, leftover_size):
            if chunk[shift] != boundary[intersection_size + shift]:
                break
        
        else:
            # Mark all as processed
            return leftover_size, None
    
    return -1, continued_intersection_sizes


def _continue_intersection_sizes(chunk, boundary, intersection_sizes):
    """
    Continues intersection sizes on a new chunk.
    
    Parameters
    ----------
    chunk : `bytes`
        Data chunk to process. Must have length > 0.
    
    boundary : `bytes`
        Boundary to check intersection with.
    
    intersection_sizes : `list<int>`
        Detected the non matched intersection sizes.
    
    Returns
    -------
    intersection_sizes : `None | list<int>`
        The new matched intersection sizes.
    """
    chunk_length = len(chunk)
    continued_intersection_sizes = None
    
    for intersection_size in intersection_sizes:
        for shift in range(0, chunk_length):
            if chunk[shift] != boundary[intersection_size + shift]:
                break
        
        else:
            if continued_intersection_sizes is None:
                continued_intersection_sizes = []
            
            continued_intersection_sizes.append(intersection_size + shift + 1)
    
    return continued_intersection_sizes


def _get_released_from_held_back(held_back, amount_to_keep):
    """
    Releases from held back chunks.
    
    Parameters
    ----------
    held_back : `None | list<bytes | memoryview>`
        The held back chunks.
    
    amount_to_keep : `int`
        The amount of bytes to hold back.
    
    Returns
    -------
    released : `None | list<bytes | memoryview>`
    held_back : `None | list<bytes | memoryview>`
    """
    if held_back is None:
        return None, None
    
    if amount_to_keep <= 0:
        return held_back, None
    
    amount_to_release = sum(len(chunk) for chunk in held_back) - amount_to_keep
    if amount_to_release <= 0:
        return None, held_back
    
    released = []
    
    while True:
        chunk = held_back[0]
        chunk_length = len(chunk)
        
        amount_to_release -= chunk_length
        if amount_to_release > 0:
            del held_back[0]
            released.append(chunk)
            
            if not held_back:
                held_back = None
                break
            
            continue
        
        if amount_to_release < 0:
            split_at = len(chunk) + amount_to_release
            chunk = memoryview(chunk)
            held_back[0] = chunk[split_at :]
            released.append(chunk[: split_at])
            break
        
        del held_back[0]
        released.append(chunk)
        if held_back:
            break
        
        held_back = None
        break
    
    return released, held_back


def _merge_intersection_sizes(intersection_sizes_0, intersection_sizes_1):
    """
    Merges two intersection sizes.
    
    Parameters
    ----------
    intersection_sizes_0 : `None | list<int>`
        Intersection sizes to merge.
    
    intersection_sizes_1 : `None | list<int>`
        Intersection sizes to merge.
    
    Returns
    -------
    intersection_sizes : `None | list<int>``
    """
    if intersection_sizes_0 is None:
        return intersection_sizes_1
    
    if intersection_sizes_1 is None:
        return intersection_sizes_0
    
    return [*intersection_sizes_0, *intersection_sizes_1]


class ReadProtocolBase(AbstractProtocolBase):
    """
    Asynchronous read protocol implementation.
    
    Scarlet-io backend uses optimistic generator based chunked readers, which have really long and complicated
    implementation, but their speed is pretty good.
    
    Attributes
    ----------
    _at_eof : `bool`
        Whether the protocol received end of file.
    
    _chunks : `Deque` of `bytes`
        Right feed, left pop queue, used to store the received data chunks.
    
    _exception : `None`, `BaseException`
        Exception set by ``.set_exception``, when an unexpected exception occur meanwhile reading from socket.
    
    _loop : ``EventThread``
        The event loop to what the protocol is bound to.
    
    _offset : `int`
        Byte offset, of the used up data of the most-left chunk.
    
    _paused : `bool`
        Whether the protocol's respective transport's reading is paused. Defaults to `False`.
        
        Also note, that not every transport supports pausing.
    
    _payload_stream : `None | Stream``
        Payload stream of the protocol.
    
    _payload_reader : `None | GeneratorType | CoroutineType`
        Payload reader generator, what gets the control back, when data, eof or any exception is received.
    
    _transport : `None | AbstractTransportLayerBase`
        Asynchronous transport implementation. Is set meanwhile the protocol is alive.
    """
    __slots__ = (
        '_at_eof', '_chunks', '_exception', '_loop', '_offset', '_paused', '_payload_reader', '_payload_stream',
        '_transport'
    )
    
    def __new__(cls, loop):
        """
        Creates a new read protocol instance.
        
        Parameters
        ----------
        loop : ``EventThread``
            The respective event loop, what the protocol uses for it's asynchronous tasks.
        """
        self = object.__new__(cls)
        self._at_eof = False
        self._loop = loop
        self._chunks = Deque()
        self._exception = None
        self._offset = 0
        self._paused = False
        self._payload_reader = None
        self._payload_stream = None
        self._transport = None
        return self
    
    
    @property
    def _payload_waiter(self):
        """
        Deprecated and will be removed at 2025 September. Please use `._payload_stream` instead.
        """
        warn(
            f'f`{type(self).__name__}._payload_stream` is deprecated and will be removed in 2025 September. '
            f'Please use `._payload_stream` instead.'
        )
        return self._payload_stream
    
    
    def __repr__(self):
        """Returns the transport's representation."""
        repr_parts = ['<', type(self).__name__]
        
        if self._at_eof:
            repr_parts.append(' at eof')
            field_added = True
        else:
            field_added = False
        
        transport = self._transport
        if (transport is not None):
            if field_added:
                repr_parts.append(', ')
            else:
                field_added = True
            
            repr_parts.append(' transport = ')
            repr_parts.append(repr(transport))
        
        exception = self._exception
        if (exception is not None):
            if field_added:
                repr_parts.append(', ')
            else:
                field_added = True
            
            repr_parts.append(' exception = ')
            repr_parts.append(repr(exception))
        
        payload_reader = self._payload_reader
        if (payload_reader is not None):
            if field_added:
                repr_parts.append(', ')
            
            repr_parts.append(' payload_reader = ')
            repr_parts.append(repr(payload_reader))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(AbstractProtocolBase.connection_made)
    def connection_made(self, transport):
        self._transport = transport
    
    
    @copy_docs(AbstractProtocolBase.connection_lost)
    def connection_lost(self, exception):
        """
        Called when the connection is lost or closed.
        
        Parameters
        ----------
        exception : `None`, `BaseException`
            Defines whether the connection is closed, or an exception was received.
            
            If the connection was closed, then `exception` is given as `None`. This can happen at the case, when eof is
            received as well.
        """
        if exception is None:
            self.eof_received()
        else:
            self.set_exception(exception)
    
    
    @copy_docs(AbstractProtocolBase.get_extra_info)
    def get_extra_info(self, name, default = None):
        transport = self._transport
        if transport is None:
            info = default
        else:
            info = transport.get_extra_info(name, default)
        
        return info
    
    
    @copy_docs(AbstractProtocolBase.close)
    def close(self):
        transport = self._transport
        if (transport is not None):
            transport.close()
    
    
    @copy_docs(AbstractProtocolBase.close_transport)
    def get_transport(self):
        return self._transport
    
    
    # read related
    def get_size(self):
        """
        Returns the received and not yet processed data's size by the protocol.
        
        Returns
        -------
        size : `int`
        """
        size = -self._offset
        for chunk in self._chunks:
            size += len(chunk)
        
        payload_stream = self._payload_stream
        if (payload_stream is not None):
            size += payload_stream.get_buffer_size()
        
        return size
    
    
    def is_at_eof(self):
        """
        Returns whether the protocol is at eof. If at eof, but has content left, then returns `False` however.
        
        Returns
        -------
        at_eof : `bool`
        """
        if not self._at_eof:
            return False
        
        # Note:
        # we cannot have `payload_reader` set because that consumes all of our `size` at once.
        
        if self.get_size():
            return False
        
        return True
    
    
    @copy_docs(AbstractProtocolBase.set_exception)
    def set_exception(self, exception):
        self._exception = exception
        
        payload_stream = self._payload_stream
        if (payload_stream is None):
            return
        
        self._payload_stream = None
        payload_stream.set_done_exception(exception)
        
        self._payload_reader.close()
        self._payload_reader = None
    
    
    @copy_docs(AbstractProtocolBase.eof_received)
    def eof_received(self):
        self._at_eof = True
        payload_reader = self._payload_reader
        if payload_reader is None:
            return False
        
        payload_stream = self._payload_stream
        self._payload_reader = None
        self._payload_stream = None
        
        try:
            payload_reader.throw(EOFError(b''))
        except EOFError as exception:
            new_exception = ConnectionError('Connection closed unexpectedly with EOF.')
            new_exception.__cause__ = exception
            payload_stream.set_done_exception(new_exception)
        
        except StopIteration:
            payload_stream.set_done_success()
        
        except GeneratorExit as exception:
            new_exception = ConnectionError('Payload reader destroyed.')
            new_exception.__cause__ = exception
            payload_stream.set_done_exception(new_exception)
        
        except BaseException as exception:
            payload_stream.set_done_exception(exception)
        
        else:
            payload_reader.close()
            new_exception = RuntimeError('Payload stream ignored eof.')
            payload_stream.set_done_exception(new_exception)
        
        return False
    
    
    @copy_docs(AbstractProtocolBase.data_received)
    def data_received(self, data):
        if not data:
            return
        
        payload_reader = self._payload_reader
        if (payload_reader is None):
            chunks = self._chunks
            chunks.append(data)
        else:
            try:
                payload_reader.send(data)
            except BaseException as exception:
                payload_stream = self._payload_stream
                self._payload_reader = None
                self._payload_stream = None
                
                if isinstance(exception, StopIteration):
                    payload_stream.set_done_success()
                
                elif isinstance(exception, EOFError):
                    new_exception = ConnectionError('Connection closed unexpectedly with EOF.')
                    new_exception.__cause__ = exception
                    payload_stream.set_done_exception(new_exception)
                
                elif isinstance(exception, GeneratorExit):
                    new_exception = ConnectionError('Payload reader destroyed.')
                    new_exception.__cause__ = exception
                    payload_stream.set_done_exception(new_exception)
                
                
                else:
                    payload_stream.set_done_exception(exception)
        
        # Pause if buffer is too big
        self._pause_reading()
    
    
    def _pause_reading(self):
        """
        Called when the protocol should consider pausing reading of its transport.
        """
        if (self.get_size() > BUFFER_SIZE_MAX) and (not self._paused):
            transport = self._transport
            if (transport is not None):
                try:
                    transport.pause_reading()
                except (AttributeError, NotImplementedError):
                    # cant be paused
                    self._transport = None
                else:
                    self._paused = True
    
    
    def _resume_reading(self):
        """
        Called when the protocol should consider resuming reading of its transport.
        """
        if self._paused:
            self._paused = False
            self._transport.resume_reading()
    
    
    def cancel_current_reader(self):
        """
        Cancels the current reader task of the protocol.
        
        Notes
        -----
        The used data by the payload reader task cannot be reused.
        """
        payload_reader = self._payload_reader
        if payload_reader is None:
            return
        
        self._payload_reader = None
        
        payload_stream = self._payload_stream
        self._payload_stream = None
        payload_stream.set_done_cancelled()
        
        payload_reader.close()
    
    
    def set_payload_reader(self, payload_reader_function):
        """
        Sets payload reader to the protocol.
        
        Parameters
        ----------
        payload_reader_function : `callable`
            A function, that gets control, every time a chunk is received.
        
        Returns
        -------
        payload_stream : ``PayloadStream``
        """
        assert self._payload_reader is None, (
            f'Payload reader already set! '
            f'payload_reader = {self._payload_reader!r}; payload_reader_function = {payload_reader_function!r}'
        )
        
        payload_stream = PayloadStream(self)
        
        exception = self._exception
        if (exception is not None):
            payload_stream.set_done_exception(exception)
        
        else:
            payload_reader = payload_reader_function(payload_stream)
            
            try:
                payload_reader.send(None)
            except StopIteration:
                payload_stream.set_done_success()
            
            except EOFError as exception:
                new_exception = ConnectionError('Connection closed unexpectedly with EOF.')
                new_exception.__cause__ = exception
                payload_stream.set_done_exception(new_exception)
            
            except GeneratorExit as exception:
                new_exception = ConnectionError('Payload reader destroyed.')
                new_exception.__cause__ = exception
                payload_stream.set_done_exception(new_exception)
            
            except BaseException as exception:
                payload_stream.set_done_exception(exception)
            
            else:
                self._payload_stream = payload_stream
                self._payload_reader = payload_reader
        
        return payload_stream
    
    
    def handle_payload_stream_abortion(self):
        """
        If you expect, that the payload waiter will be cancelled from outside, call this method to throw eof into the
        protocol at that case.
        """
        payload_stream = self._payload_stream
        if (payload_stream is not None):
            payload_stream.add_done_callback(self._payload_stream_abortion_callback)
    
    
    def _payload_stream_abortion_callback(self, payload_stream):
        """
        Callback added to ``._payload_stream`` by ``.handle_payload_stream_abortion`` to throw eof into the
        ``.payload_reader`` task if payload stream is aborted from outside.
        
        Parameters
        ----------
        payload_stream : ``PayloadStream``
            The respective ``._payload_stream``.
        """
        if payload_stream.is_aborted():
            self.eof_received()
    
    
    async def read(self, n = -1):
        """
        Reads up to `n` amount of bytes from the protocol.
        
        If `n` is given as a negative integer, then will read until eof.
        
        This method is a coroutine.
        
        Parameters
        ----------
        n : `int`
            The amount of bytes to read. Defaults to `-1`.
        
        Returns
        -------
        result : `bytes`
        """
        if n < 0:
            payload_stream = self.set_payload_reader(self._read_until_eof)
        
        elif n > 0:
            payload_stream = self.set_payload_reader(partial_func(self._read_exactly, n, True))
        
        else:
            exception = self._exception
            if (exception is not None):
                raise exception
            
            return b''
        
        return (await payload_stream)
    
    
    async def read_exactly(self, n):
        """
        Read exactly `n` bytes from the protocol.
        
        This method is a coroutine.
        
        Parameters
        ----------
        n : `int`
            The amount of bytes to read. Cannot be negative.
        
        Returns
        -------
        result : `str`
        
        Raises
        ------
        ValueError
            If `n` is less than `0`.
        ConnectionError
            Connection lost before `n` bytes were received.
        """
        
        if n > 0:
            return (await self.set_payload_reader(partial_func(self._read_exactly, n, False)))
        
        if n < 0:
            raise ValueError(f'`.read_exactly` size can not be less than `0`, got {n!r}.')
        
        # n == 0
        exception = self._exception
        if (exception is not None):
            raise exception
        
        return b''
    
    
    async def read_line(self):
        raise NotImplementedError
    
    
    async def read_until(self, boundary):
        """
        Reads until the given boundary.
        This method is a coroutine.
        
        Parameters
        ----------
        boundary : `bytes`
            The boundary to read until. Consumed but not returned.
        
        Raises
        ------
        ConnectionError
            Connection lost before boundary hit.
        """
        if not boundary:
            exception = self._exception
            if (exception is not None):
                raise exception
            
            return b''
        
        return await self.set_payload_reader(partial_func(self._read_until, boundary))
    
    
    async def read_once(self):
        """
        Waits till exactly one chunk is of data is received.
        
        This method is a coroutine.
        
        Returns
        -------
        data : `bytes`
        
        Raises
        ------
        ConnectionError
            At end of file.
        """
        return (await self.set_payload_reader(self._read_once))
    
    
    @to_coroutine
    def _wait_for_data(self):
        """
        Payload reader task helper, what waits for 1 chunk to be receive, then adds it to ``._chunks`` and also returns
        it as well.
        
        This method is an awaitable generator.
        
        Returns
        -------
        chunk : `bytes`
        """
        self._resume_reading()
        
        chunk = yield
        self._chunks.append(chunk)
        return chunk
    
    
    async def _read_exactly_by_chunk(self, n):
        """
        Reads exactly `n` chunks and yields them back.
        
        This method is a coroutine generator.
        
        Parameters
        ----------
        n : `int`
            The amount of bytes to read. Must be greater than 0.
        
        Returns
        -------
        chunk : `bytes-like`
        
        Raises
        ------
        EofError
            Connection lost before `n` bytes were received.
        """
        chunks = self._chunks
        if chunks:
            chunk = chunks[0]
            offset = self._offset
        else:
            if self._at_eof:
                raise EOFError(b'')
            
            chunk = await self._wait_for_data()
            offset = 0
        
        chunk_size = len(chunk)
        if offset == 0:
            if chunk_size > n:
                self._offset = n
                yield memoryview(chunk)[:n]
                return
            
            # chunk same size as the requested?
            if chunk_size == n:
                del chunks[0]
                # offset is already 0, nice!
                yield chunk
                return
            
            n -= len(chunk)
            del chunks[0]
            yield chunk
        
        else:
            end = offset + n
            if chunk_size > end:
                self._offset = end
                yield memoryview(chunk)[offset : end]
                return
            
            # chunk_size + offset end when the requested's end is.
            if chunk_size == end:
                del chunks[0]
                self._offset = 0
                yield memoryview(chunk)[offset:]
                return
            
            n -= (chunk_size - offset)
            del chunks[0]
            yield memoryview(chunk)[offset:]
        
        while True:
            if chunks:
                chunk = chunks[0]
            else:
                if self._at_eof:
                    self._offset = 0
                    raise EOFError(b'')
                
                chunk = await self._wait_for_data()
            
            chunk_size = len(chunk)
            
            n -= chunk_size
            if n > 0:
                yield chunk
                del chunks[0]
                continue
            
            if n == 0:
                yield chunk
                del chunks[0]
                self._offset = 0
                return
            
            offset = self._offset = chunk_size + n
            yield memoryview(chunk)[:offset]
            return
    
    
    async def _read_exactly(self, n, allow_eof, payload_stream):
        """
        Payload reader task, what reads exactly `n` bytes from the protocol.
        
        This method is a generator.
        
        Parameters
        ----------
        n : `int`
            The amount of bytes to read. Must be greater than 0.
        
        allow_eof : `bool`
            Whether end of file is allowed.
        
        payload_stream : ``PayloadStream``
            Payload buffer to read into.
        
        Raises
        ------
        EofError
            Connection lost before `n` bytes were received.
        """
        try:
            async for chunk in self._read_exactly_by_chunk(n):
                payload_stream.add_received_chunk(chunk)
        except EOFError:
            if not allow_eof:
                raise
        
        payload_stream.set_done_success()
    
    
    async def _read_exactly_as_one(self, n):
        """
        Reads exactly `n` bytes of data at once and returns it. Can be used by other payload readers.
        
        This method is a coroutine.
        
        Parameters
        ----------
        n : `int`
            The amount of bytes to read. Must be greater than 0.
        
        Raises
        ------
        EofError
            Connection lost before `n` bytes were received.
        """
        chunks = []
        async for chunk in self._read_exactly_by_chunk(n):
            chunks.append(chunk)
        
        return b''.join(chunks)
    
    
    async def _read_until_by_chunk(self, boundary):
        """
        Payload reader task, what reads until `boundary` is hit. Yields back each data chunk.
        
        This method is a coroutine.
        
        Parameters
        ----------
        boundary : `bytes`
            The boundary to read until. Consumed but not returned.
        
        Raises
        ------
        EofError
            Connection lost before boundary hit.
        """
        chunks = self._chunks
        intersection_sizes = None
        held_back = None
        
        while True:
            if chunks:
                chunk = chunks[0]
                offset = self._offset
            else:
                if self._at_eof:
                    raise EOFError(b'')
                
                chunk = await self._wait_for_data()
                offset = 0
            
            if (intersection_sizes is not None):
                offset, intersection_sizes = _finish_intersection_sizes(chunk, boundary, intersection_sizes)
                # Found boundary in between?
                if offset != -1:
                    # Do not offset if offset if at the end of the chunk. Delete chunk instead.
                    if offset == len(chunk):
                        del chunks[0]
                        offset = 0
                    self._offset = offset
                    
                    released, held_back = _get_released_from_held_back(held_back, len(boundary) - offset)
                    if (released is not None):
                        for to_release in released:
                            yield to_release
                    released = None
                    return
                
                if (intersection_sizes is None):
                    # Release all.
                    for to_release in held_back:
                        yield to_release
                    held_back = None
                    self._offset = 0
                    offset = 0
                
                else:
                    del chunks[0]
                    self._offset = 0
                    # Chunk too small?
                    held_back.append(chunk)
                    intersection_sizes = _merge_intersection_sizes(
                        _continue_intersection_sizes(chunk, boundary, intersection_sizes),
                        _get_end_intersection_sizes(chunk, 0, boundary),
                    )
                    released, held_back = _get_released_from_held_back(held_back, len(boundary))
                    if (released is not None):
                        for to_release in released:
                            yield to_release
                    
                    released = None
                    continue
            
            index = chunk.find(boundary, offset)
            
            # Found boundary?
            if index != -1:
                original_offset = offset
                offset = index + len(boundary)
                if offset == len(chunk):
                    del chunks[0]
                    offset = 0
                self._offset = offset
                
                if original_offset != index:
                    yield memoryview(chunk)[original_offset : index]
                return
            
            del chunks[0]
            self._offset = 0
            if offset:
                chunk = memoryview(chunk)[offset:]
            intersection_sizes = _get_end_intersection_sizes(chunk, offset, boundary)
            if (intersection_sizes is not None):
                offset = len(chunk) - intersection_sizes[0]
                held_back = [memoryview(chunk)[offset:]]
                
                # Do not yield if `offset == 0` 
                if not offset:
                    continue
                
                chunk = memoryview(chunk)[:offset]
            
            yield chunk
            
            # Repeat loop
            continue
    
    
    async def _read_until(self, boundary, payload_stream):
        """
        Payload reader task, what reads until `boundary` is hit.
        
        This method is a coroutine.
        
        Parameters
        ----------
        boundary : `bytes`
            The boundary to read until. Consumed but not returned.
        
        payload_stream : ``PayloadStream``
            Payload buffer to read into.
        
        Raises
        ------
        EofError
            Connection lost before boundary hit.
        CancelledError
            If the reader task is cancelled not by receiving eof.
        """
        async for chunk in self._read_until_by_chunk(boundary):
            payload_stream.add_received_chunk(chunk)
    
    
    async def _read_until_as_one(self, boundary):
        """
        Reads until `boundary` is hit at once and returns it. Can be used by other payload readers.
        
        This method is a coroutine.
        
        Parameters
        ----------
        boundary : `bytes`
            The boundary to read until. Consumed but not returned.
        
        Raises
        ------
        EofError
            Connection lost before boundary hit.
        """
        chunks = []
        async for chunk in self._read_until_by_chunk(boundary):
            chunks.append(chunk)
        
        return b''.join(chunks)
    
    
    async def _read_until_eof(self, payload_stream):
        """
        Payload reader task, which reads the protocol until EOF is hit.
        
        This method is a coroutine.
        
        Parameters
        ----------
        payload_stream : ``PayloadStream``
            Payload buffer to read into.
        """
        chunks = self._chunks
        offset = self._offset
        self._offset = 0
        while chunks:
            chunk = chunks.popleft()
            if offset:
                chunk = memoryview(chunk)[offset :]
                offset = 0
            
            payload_stream.add_received_chunk(chunk)
        
        while not self._at_eof:
            try:
                chunk = await self._wait_for_data()
            except EOFError:
                break
            
            except GeneratorExit:
                payload_stream.set_done_exception(ConnectionError('Payload reader destroyed.'))
                raise
            
            del chunks[0]
            payload_stream.add_received_chunk(chunk)
        
        payload_stream.set_done_success()
    
    
    async def _read_once(self, payload_stream):
        """
        Reader task for reading exactly one chunk out.
        
        This method is a coroutine.
        
        Parameters
        ----------
        payload_stream : ``PayloadStream``
            Payload buffer to read into.
        """
        chunks = self._chunks
        offset = self._offset
        if chunks:
            self._offset = 0
            chunk = chunks.popleft()
            if offset:
                chunk = memoryview(chunk)[offset :]
            
            payload_stream.add_received_chunk(chunk)
            payload_stream.set_done_success()
            return
        
        if self._at_eof:
            return
        
        try:
            chunk = await self._wait_for_data()
        except EOFError:
            pass
        
        else:
            del chunks[0]
            payload_stream.add_received_chunk(chunk)


class ReadWriteProtocolBase(ReadProtocolBase):
    """
    Read-write asynchronous protocol implementation.
    
    Attributes
    ----------
    _at_eof : `bool`
        Whether the protocol received end of file.
    
    _chunks : `Deque` of `bytes`
        Right feed, left pop queue, used to store the received data chunks.
    
    _exception : `None`, `BaseException`
        Exception set by ``.set_exception``, when an unexpected exception occur meanwhile reading from socket.
    
    _loop : ``EventThread``
        The event loop to what the protocol is bound to.
    
    _offset : `int`
        Byte offset, of the used up data of the most-left chunk.
    
    _paused : `bool`
        Whether the protocol's respective transport's reading is paused. Defaults to `False`.
        
        Also note, that not every transport supports pausing.
    
    _payload_reader : `None`, `GeneratorType`
        Payload reader generator, what gets the control back, when data, eof or any exception is received.
    
    _payload_stream : `None | PayloadStream`
        Payload stream of the protocol.
    
    _transport : `None`, `object`
        Asynchronous transport implementation. Is set meanwhile the protocol is alive.
    
    _drain_waiter : `None`, ``Future``
        A future, what is used to block the writing task, till it's writen data is drained.
    """
    __slots__ = ('_drain_waiter', )
    
    def __new__(cls, loop):
        """
        Creates a new read-write protocol instance.
        
        Parameters
        ----------
        loop : ``EventThread``
            The respective event loop, what the protocol uses for it's asynchronous tasks.
        """
        self = ReadProtocolBase.__new__(cls, loop)
        
        self._drain_waiter = None
        
        return self
    
    
    @copy_docs(AbstractProtocolBase.pause_writing)
    def pause_writing(self):
        self._paused = True
    
    
    @copy_docs(AbstractProtocolBase.resume_writing)
    def resume_writing(self):
        self._paused = False
        
        drain_waiter = self._drain_waiter
        if drain_waiter is None:
            return
        
        self._drain_waiter = None
        drain_waiter.set_result_if_pending(None)
    
    
    async def _drain_helper(self):
        """
        Called when the transport buffer after writing goes over the high limit mark to wait till it is drained.
        
        This method is a coroutine.
        """
        if not self._paused:
            return
        
        drain_waiter = Future(self._loop)
        self._drain_waiter = drain_waiter
        await drain_waiter
    
    
    @copy_docs(ReadProtocolBase.connection_lost)
    def connection_lost(self, exception):
        if exception is None:
            self.eof_received()
        else:
            self.set_exception(exception)
        
        # wake up the writer if currently paused.
        if not self._paused:
            return
        
        drain_waiter = self._drain_waiter
        if drain_waiter is None:
            return
        
        self._drain_waiter = None
        if drain_waiter.is_done():
            return
        
        if exception is None:
            drain_waiter.set_result(None)
        else:
            drain_waiter.set_exception(exception)
    
    
    @copy_docs(AbstractProtocolBase.write)
    def write(self, data):
        transport = self._transport
        if transport is None:
            raise RuntimeError(f'Protocol has no attached transport; self={self!r}.')
        
        transport.write(data)
    
    
    @copy_docs(AbstractProtocolBase.writelines)
    def writelines(self, lines):
        transport = self._transport
        if transport is None:
            raise RuntimeError(f'Protocol has no attached transport; self={self!r}.')
        
        transport.writelines(lines)
    
    
    @copy_docs(AbstractProtocolBase.write_eof)
    def write_eof(self):
        transport = self._transport
        if (transport is not None):
            transport.write_eof()
    
    
    @copy_docs(AbstractProtocolBase.can_write_eof)
    def can_write_eof(self):
        transport = self._transport
        if (transport is None):
            return False
        
        return transport.can_write_eof()
    
    
    @copy_docs(AbstractProtocolBase.drain)
    async def drain(self):
        # use after writing
        exception = self._exception
        if (exception is not None):
            raise exception
        
        transport = self._transport
        if (transport is not None):
            if transport.is_closing():
                # skip 1 loop, so connection_lost() will be called
                await skip_ready_cycle()
        
        await self._drain_helper()



class DatagramAddressedReadProtocol(AbstractProtocolBase):
    """
    Datagram reader protocol for reading from payloads form multiple addresses.
    
    Attributes
    ----------
    _by_address : `dict` of (`tuple` (`str`, `int`), ``ReadProtocolBase``) items
        Dictionary to store the alive readers by address.
    _loop : ``EventThread``
        The loop to what the protocol is bound to.
    _protocol_factory : `type`
        Protocol type to create for each address.
    _transport : `object`
        Asynchronous transport implementation, what calls the protocol's ``.datagram_received`` when data is
        received.
    _waiters : `None`, `list` of `Future`
        Waiters for any payload receive if applicable.
    """
    __slots__ = ('_by_address', '_loop', '_protocol_factory', '_transport', '_waiters',)
    
    def __new__(cls, loop, protocol_factory):
        """
        Creates a new datagram addressed read protocol instance bound to the given loop.
        
        Parameters
        ----------
        loop : ``EventThread``
            The loop to what the protocol gonna be bound to.
        protocol_factory : `callable`
            Protocol type to create for each address.
            
            Defaults to ``ReadProtocolBase``.
        """
        self = object.__new__(cls)
        self._loop = loop
        self._by_address = {}
        self._transport = None
        self._waiters = None
        self._protocol_factory = protocol_factory
        return self
    
    
    @copy_docs(AbstractProtocolBase.connection_made)
    def connection_made(self, transport):
        self._transport = transport
        for reader_protocol in self._by_address.values():
            reader_protocol.connection_made(transport)
    
    
    @copy_docs(AbstractProtocolBase.connection_lost)
    def connection_lost(self, exception):
        for reader_protocol in self._by_address.values():
            if exception is None:
                reader_protocol.eof_received()
            else:
                reader_protocol.set_exception(exception)
    
    
    @copy_docs(AbstractProtocolBase.datagram_received)
    def datagram_received(self, data, address):
        by_address = self._by_address
        try:
            protocol = by_address[address]
        except KeyError:
            protocol = self._protocol_factory()
            protocol.connection_made(self._transport)
            by_address[address] = protocol
        
        protocol.data_received(data)
        
        waiters = self._waiters
        if (waiters is not None):
            self._waiters = None
            result = (address, protocol)
            for waiter in waiters:
                waiter.set_result_if_pending(result)
    
    
    @copy_docs(AbstractProtocolBase.error_received)
    def error_received(self, exception):
        pass
    
    
    async def wait_for_receive(self, address = None, timeout = None):
        """
        Can be used to wait for payload to receive. Note, that this method should be used only initially, because the
        reader protocols implement the reading.
        
        This method is a coroutine.
        
        Parameters
        ----------
        address : `None`, `tuple` (`str`, `int`) = `None`, Optional
            The address of which payload is waiter for.
        timeout : `None`, `float = `None`, Optional
            The maximal amount of time to wait before raising `TimeoutError`.
        
        Returns
        -------
        result : `None`, ``ReadProtocolBase`` or (`tuple` (`str`, `int`), ``ReadProtocolBase``)
            - If `timeout` is given and timeout occur, then returns `None`.
            - if `address` is given and data is received from ir, then returns the respective ``ReadProtocolBase``.
            - If `address` is not given, then returns a `tuple` of the respective `address` and protocol.
        
        Raises
        ------
        TimeoutError
            - If timeout occurred.
        """
        if timeout is None:
            if address is None:
                waiters = self._waiters
                if waiters is None:
                    self._waiters = waiters = []
                
                waiter = Future(self._loop)
                waiters.append(waiter)
                
                return await waiter
            else:
                while True:
                    waiters = self._waiters
                    if waiters is None:
                        self._waiters = waiters = []
                    
                    waiter = Future(self._loop)
                    waiters.append(waiter)
                    
                    address, protocol = await waiter
                    if address == address:
                        return protocol
                    
                    continue
        
        else:
            waiter = Task(self._loop, self.wait_for_receive(address))
            waiter.apply_timeout(timeout)
            
            try:
                result = await waiter
            except TimeoutError:
                result = None
            
            return result
    
    
    @copy_docs(AbstractProtocolBase.close)
    def close(self):
        transport = self._transport
        if (transport is not None):
            transport.close()
    
    
    @copy_docs(AbstractProtocolBase.close_transport)
    def get_transport(self):
        return self._transport


class DatagramMergerReadProtocol(ReadProtocolBase):
    """
    Asynchronous read protocol implementation.
    
    Scarlet-io backend uses optimistic generator based chunked readers, which have really long and complicated
    implementation, but their speed is pretty good.
    
    Attributes
    ----------
    _at_eof : `bool`
        Whether the protocol received end of file.
    _chunks : `Deque` of `bytes`
        Right feed, left pop queue, used to store the received data chunks.
    _exception : `None`, `BaseException`
        Exception set by ``.set_exception``, when an unexpected exception occur meanwhile reading from socket.
    _loop : ``EventThread``
        The event loop to what the protocol is bound to.
    _offset : `int`
        Byte offset, of the used up data of the most-left chunk.
    _paused : `bool`
        Whether the protocol's respective transport's reading is paused. Defaults to `False`.
        
        Also note, that not every transport supports pausing.
    _payload_reader : `None`, `GeneratorType`
        Payload reader generator, what gets the control back, when data, eof or any exception is received.
    _payload_stream : `None` of ``Future``
        Payload waiter of the protocol, what's result is set, when the ``.payload_reader`` generator returns.
        
        If cancelled or marked by done or any other methods, the payload reader will not be cancelled.
    _transport : `None`, `object`
        Asynchronous transport implementation. Is set meanwhile the protocol is alive.
    """
    __slots__ = ()
    
    @copy_docs(AbstractProtocolBase.datagram_received)
    def datagram_received(self, data, address):
        self.data_received(data)



__all__ = ()


from ...weak_value_dictionary import WeakValueDictionary

from collections import OrderedDict


CONSOLE_INPUT_CACHE = OrderedDict()
CONSOLE_INPUT_MAX_SIZE = 1000000
CONSOLE_INPUT_CACHE_OUTFLOW_CHECK = 17

FILE_INFO_CACHE = WeakValueDictionary()

LINE_INFO_CACHE = OrderedDict()
LINE_CACHE_MAX_SIZE = 1000

LINE_CACHE_SESSIONS = set()



__all__ = ()

import sys

from functools import partial as partial_func
from os.path import sep as PATH_SEPARATOR
from reprlib import repr as short_repr
from types import AsyncGeneratorType as CoroutineGeneratorType, CoroutineType, GeneratorType

from ..function_tools import alchemy_incendiary


def _cut_file_name(file_name):
    """
    Cuts the given file name.
    
    Parameters
    ----------
    file_name : `None | str`
        The file name to cut.
    
    Returns
    -------
    file_name : `None | str`
    """
    if (file_name is None) or (not file_name):
        return None
    
    file_name_length = len(file_name)
    
    for path in sys.path:
        path_length = len(path)
        
        if not file_name.startswith(path):
            continue
        
        if file_name_length <= path_length:
            continue
        
        if file_name[path_length] != PATH_SEPARATOR:
            continue
        
        return '...' + file_name[path_length:]
    
    return file_name


def _merge_nullable_tuples(tuple_0, tuple_1):
    """
    Merges two nullable tuple.
    
    Parameters
    ----------
    tuple_0 : `None | tuple`
        Tuple to merge.
    tuple_1 : `None | tuple`
        Tuple ot merge.
    
    Returns
    -------
    merged : `None | tuple`
    """
    if tuple_0 is None:
        return tuple_1
    
    if tuple_1 is None:
        return tuple_0
    
    return tuple_0 + tuple_1


def _merge_nullable_dicts(dict_0, dict_1):
    """
    Merges two nullable tuple.
    
    Parameters
    ----------
    dict_0 : `None | dict<str, object>`
        Dictionary to merge.
    dict_1 : `None | dict<str, object>`
        Dictionary ot merge.
    
    Returns
    -------
    merged : `None | dict<str, object>`
    """
    if dict_0 is None:
        return dict_1
    
    if dict_1 is None:
        return dict_0
    
    return {**dict_1, **dict_0}


def format_callback(function, positional_parameters = None, keyword_parameters = None):
    """
    Formats the given callback to a more user friendly representation.
    
    Parameters
    ----------
    function : `callable`
        The callback to format.
    positional_parameters : `None | tuple<object>` = `None`, Optional
        Additional parameters to call the `function` with.
    keyword_parameters : `None | dict<str, object>` = `None`, Optional
        Additional keyword parameters to call the `function` with.
    
    Returns
    -------
    result : `str`
        The formatted callback.
    """
    # Unwrap
    while True:
        if isinstance(function, alchemy_incendiary):
            # `alchemy_incendiary` is not accepting nesting, yeet old parameters.
            positional_parameters = function.positional_parameters
            keyword_parameters = function.keyword_parameters
            function = function.function
            continue
        
        if isinstance(function, partial_func):
            # `partial_func` accepts nesting, extend old parameters.
            positional_parameters = _merge_nullable_tuples(function.args, positional_parameters)
            keyword_parameters = _merge_nullable_dicts(keyword_parameters, function.keywords)
            function = function.func
            continue
        
        break

    function_representation = getattr(function, '__qualname__', None)
    if function_representation is None:
        function_representation = getattr(function, '__name__', None)
        if function_representation is None:
            function_representation = type(function).__name__
    
    
    parts = [function_representation, '(']
    parameters_added = False
    
    # Add positional parameters
    if (positional_parameters is not None) and positional_parameters:
        parameters_added = True
        
        index = 0
        length = len(positional_parameters)
        
        while True:
            value = positional_parameters[index]
            parts.append(short_repr(value))
            
            index += 1
            if index >= length:
                break
            
            parts.append(', ')
            continue
    
    
    # Add keyword parameters
    if (keyword_parameters is not None) and keyword_parameters:
        if parameters_added:
            parts.append(', ')
        
        index = 0
        length = len(keyword_parameters)
        items = sorted(keyword_parameters.items())
        
        while True:
            key, value = items[index]
            parts.append(key)
            parts.append(' = ')
            parts.append(short_repr(value))
            
            index += 1
            if index >= length:
                break
            
            parts.append(', ')
            continue
    
    
    parts.append(')')
    return ''.join(parts)


def format_builtin(function):
    """
    Formats the given built-in's name.
    
    Parameters
    ----------
    function : `callable`
        The builtin to format.
    
    Returns
    -------
    result : `str`
        The formatted builtin.
    """
    # Cython or builtin
    name = getattr(function, '__qualname__', None)
    if name is None:
        name = getattr(function, '__name__', None)
        if name is None: # builtins might reach this part
            name = type(function).__name__
    
    return f'{name!s}()'


def format_coroutine(coroutine):
    """
    Formats the given coroutine to a more user friendly representation.
    
    Parameters
    ----------
    coroutine : `CoroutineType`, `GeneratorType`, `CoroutineGeneratorType`
        The coroutine to get representation of.
    
    Returns
    -------
    result : `str`
        The formatted coroutine.
    """
    coroutine_type = type(coroutine)
    if coroutine_type is GeneratorType:
        code = coroutine.gi_code
        frame = coroutine.gi_frame
        running = coroutine.gi_running
    
    elif coroutine_type is CoroutineType:
        code = coroutine.cr_code
        frame = coroutine.cr_frame
        running = coroutine.cr_running
    
    elif coroutine_type is CoroutineGeneratorType:
        code = coroutine.ag_code
        frame = coroutine.ag_frame
        running = coroutine.ag_running
    
    else:
        code = None
        frame = None
        running = -1
    
    if running == True:
        state = 'running'
    
    elif running == False:
        if frame is None:
            state = 'finished'
        
        else:
            state = 'suspended'
    
    else:
        state = 'unknown state'
    
    if running == -1:
        name = format_builtin(coroutine)
    else:
        name = format_callback(coroutine)
    
    if (code is None):
        location = None
    else:
        file_name = _cut_file_name(code.co_filename)
        if file_name is None:
            location = None
        else:
            if frame is None:
                line_number = code.co_firstlineno
            else:
                line_number = frame.f_lineno
            
            location = f'"{file_name!s}", line {line_number!s}'
    
    if location is None:    
        location = 'unknown location'
    
    return f'<{name!s} from {location!s}; {state!s}>'



__all__ = ('FrameProxyTraceback',)

from ...docs import copy_docs

from .frame_proxy_base import FrameProxyBase


class FrameProxyTraceback(FrameProxyBase):
    """
    Wraps a traceback's frame.
    
    Attributes
    ----------
    _traceback : ``TracebackType``
        The wrapped traceback frame.
    alike_count : `int`
        How much alike frames are in the same exception.
    expression_info : `None | ExpressionInfo`
        Additional expression information.
    """
    __slots__ = ('_traceback',)
    
    def __new__(cls, traceback):
        """
        Creates a new traceback frame proxy.
        
        Parameters
        ----------
        traceback : ``TracebackType``
            The traceback frame to wrap.
        """
        self = object.__new__(cls)
        self._traceback = traceback
        self.alike_count = 0
        self.expression_info = None
        return self
    
    
    @property
    @copy_docs(FrameProxyBase.builtins)
    def builtins(self):
        return self._traceback.tb_frame.f_builtins
    
    
    @property
    @copy_docs(FrameProxyBase.globals)
    def globals(self):
        return self._traceback.tb_frame.f_globals
    
    
    @property
    @copy_docs(FrameProxyBase.locals)
    def locals(self):
        return self._traceback.tb_frame.f_locals
    
    
    @property
    @copy_docs(FrameProxyBase.code)
    def code(self):
        return self._traceback.tb_frame.f_code
    
    
    @property
    @copy_docs(FrameProxyBase.instruction_index)
    def instruction_index(self):
        return self._traceback.tb_frame.f_lasti
    
    
    @property
    @copy_docs(FrameProxyBase.line_index)
    def line_index(self):
        return self._traceback.tb_lineno - 1
    
    
    @property
    @copy_docs(FrameProxyBase.tracing_function)
    def tracing_function(self):
        return self._traceback.tb_frame.f_trace
    
    
    @copy_docs(FrameProxyBase.has_variables)
    def has_variables(self):
        return True



__all__ = ('ContentType', 'ContentTypeParsingError', 'parse_content_type')

from ..utils import MultiValueDictionary, RichAttributeErrorBaseType


WHITE_SPACE = ' \t' # space & horizontal tab
DELIMITERS = (
    ',:;' # field delimiters
    '(){}[]<>\'\"' # Bracket delimiters
    '\\?@=/' # These are also not allowed
)


def _consume_whitespace(string, index, end):
    """
    Consumes all white space (and horizontal tabulator) till hits end of string or any other character.
    
    Parameters
    ----------
    string : `str`
        The string to parse.
    
    index : `int`
        The index where to start the parsing.
    
    end : `int`
        The end where to stop parsing.
    
    Returns
    -------
    index : `int` 
    """
    while True:
        if index >= end:
            break
        
        character = string[index]
        if character not in ' \t':
            break
        
        index += 1
        continue
    
    return index


def _parse_token(string, index, end, disallowed):
    """
    Parses a single token out. Stops when hits the first non-token character.
    
    Parameters
    ----------
    string : `str`
        The string to parse.
    
    index : `int`
        The index where to start the parsing.
    
    end : `int`
        The end where to stop parsing.
    
    disallowed : `None | str`
        Disallowed characters that cannot be part of then token.
    
    Returns
    -------
    index : `int` 
    token : `None | str`
    """
    start = index
    
    while True:
        if index >= end:
            break
        
        character = string[index]
        if (disallowed is not None) and (character in disallowed):
            break
        
        if (character in WHITE_SPACE) or (character in DELIMITERS):
            break
        
        index += 1
        continue
    
    if start >= index:
        token = None
    else:
        token = string[start : index]
    
    return index, token


def _parse_quoted(string, index, end):
    """
    Parses a quoted value. The start of the quoted value should be already consumed.
    Stops when the quote is closed (consuming it) or when we end of string is hit.
    
    Parameters
    ----------
    string : `str`
        The string to parse.
    
    index : `int`
        The index where to start the parsing.
    
    end : `int`
        The end where to stop parsing.
    
    Returns
    -------
    index : `int` 
    token : `None | str`
    expected : `None | str`
    """
    collected = None
    expected = None
    
    last_escape = False
    
    while True:
        if index >= end:
            if last_escape:
                if collected is None:
                    collected = []
                
                collected.append('\\')
            
            expected = '"'
            break
        
        character = string[index]
        
        index += 1
        
        if last_escape:
            last_escape = False
            
            if collected is None:
                collected = []
            
            if character not in '\\"':
                collected.append('\\')
            
            collected.append(character)
        
        else:
            if character == '"':
                break
            
            if character == '\\':
                last_escape = True
            
            else:
                if collected is None:
                    collected = []
                
                collected.append(character)
        
        continue
    
    if collected is None:
        value = None
    else:
        value = ''.join(collected)
    
    return index, value, expected


def _parse_token_with_space_around(string, index, end, disallowed):
    """
    Parses a token. Allows space to be around.
    
    Parameters
    ----------
    string : `str`
        The string to parse.
    
    index : `int`
        The index where to start the parsing.
    
    end : `int`
        The end where to stop parsing.
    
    disallowed : `None | str`
        Disallowed characters that cannot be part of then token.
    
    Returns
    -------
    index : `int` 
    token : `None | str`
    """
    index = _consume_whitespace(string, index, end)
    index, token = _parse_token(string, index, end, disallowed)
    index = _consume_whitespace(string, index, end)
    
    return index, token


def _parse_token_or_quited_with_space_around(string, index, end):
    """
    Parses a token or a quoted value out. Allows space to be around.
    
    Parameters
    ----------
    string : `str`
        The string to parse.
    
    index : `int`
        The index where to start the parsing.
    
    end : `int`
        The end where to stop parsing.
    
    Returns
    -------
    index : `int` 
    token : `None | str`
    expected : `None | str`
    """
    token = None
    expected = None
    
    index = _consume_whitespace(string, index, end)
    if index < end:
        character = string[index]
        if character == '"':
            index += 1
            index, token, expected = _parse_quoted(string, index, end)
        else:
            index, token = _parse_token(string, index, end, None)
    
    index = _consume_whitespace(string, index, end)
    return index, token, expected


def _parse_head(string, index, end):
    """
    Parses a content-type header's head.
    
    Parameters
    ----------
    string : `str`
        The string to parse.
    
    index : `int`
        The index where to start the parsing.
    
    end : `int`
        The end where to stop parsing.
    
    Returns
    -------
    index : `int`
    type_ : `None | str`
    sub_type : `None | str`
    suffix : `None | str`
    expected : `None | str`
    """
    type_ = None
    sub_type = None
    suffix = None
    expected = None
    
    while True:
        if index >= end:
            break
        
        index, type_ = _parse_token_with_space_around(string, index, end, ';/')
        if index >= end:
            break
        
        character = string[index]
        if character == ';':
            index += 1
            break
        
        if character != '/':
            expected = ';/'
            break
        
        index += 1
        index, sub_type = _parse_token_with_space_around(string, index, end, ';+')
        if index >= end:
            break
        
        character = string[index]
        if character == ';':
            index += 1
            break
        
        if character != '+':
            expected = ';+'
            break
        
        index += 1
        index, suffix = _parse_token_with_space_around(string, index, end, ';')
        if index >= end:
            break
        
        character = string[index]
        if character == ';':
            index += 1
            break
        
        expected = ';'
        break
    
    return index, type_, sub_type, suffix, expected


def _parse_parameter(string, index, end):
    """
    Parses a single parameter out. If ends with a semi colon, consumes it as well.
    
    Parameters
    ----------
    string : `str`
        The string to parse.
    
    index : `int`
        The index where to start the parsing.
    
    end : `int`
        The end where to stop parsing.
    
    Returns
    -------
    index : `int`
    type_ : `None | str`
    sub_type : `None | str`
    suffix : `None | str`
    expected : `None | str`
    """
    key = None
    value = None
    expected = None
    
    while True:
        if index >= end:
            break
        
        index, key = _parse_token_with_space_around(string, index, end, ';=')
        if index >= end:
            break
        
        character = string[index]
        if character == ';':
            index += 1
            break
        
        if character != '=':
            expected = ';='
            break
        
        index += 1
        index, value, expected = _parse_token_or_quited_with_space_around(string, index, end)
        if index >= end:
            break
        
        if string[index] == ';':
            index += 1
            break
        
        expected = ';'
        break
    
    return index, key, value, expected


def _parse_content_type(string):
    """
    Parses the given content-type header value.
    
    Parameters
    ----------
    string : `str`
        The string to parse.
    
    Returns
    -------
    type_ : `None | str`
    sub_type : `None | str`
    suffix : `None | str`
    parameters : `None | MultiValueDictionary<str, str>`
    expected : `None | str`
    """
    index = 0
    end = len(string)
    
    parameters = None
    
    index, type_, sub_type, suffix, expected = _parse_head(string, index, end)
    
    # type, sub_type, suffix are case in-sensitive
    if (type_ is not None):
        type_ = type_.casefold()
    
    if (sub_type is not None):
        sub_type = sub_type.casefold()
    
    if (suffix is not None):
        suffix = suffix.casefold()
    
    # Convert `*` head to `*/*`
    if (type_ == '*') and (sub_type is None):
        sub_type = '*'
    
    if (expected is None) and (index < end):
        while True:
            index, key, value, expected = _parse_parameter(string, index, end)
            
            # do we have an item?
            if (key is not None) or (value is not None):
                # key must be set & must is case in-sensitive
                if (key is None):
                    key = ''
                else:
                    key = key.casefold()
                
                # value must be set
                if (value is None):
                    value = ''
                
                if (parameters is None):
                    parameters = MultiValueDictionary()
                
                parameters[key] = value
            
            if (expected is not None):
                break
            
            if (index >= end):
                break
    
    if (expected is None):
        error = None
    else:
        error = (string, index, expected)
    
    return type_, sub_type, suffix, parameters, error


def parse_content_type(string):
    """
    Parses the given content-type header value. `None` allowed.
    
    Parameters
    ----------
    string : `None | str`
        The string to parse.
    
    Returns
    -------
    content_type : ``ContentType``
    content_type_parsing_error : `None | ContentTypeParsingError`
    """
    if (string is None) or (not string):
        content_type = ContentType.create_empty()
        content_type_parsing_error = None
    
    else:
        type_, sub_type, suffix, parameters, error = _parse_content_type(string)
        content_type = ContentType(type_, sub_type, suffix, parameters)
        if (error is None):
            content_type_parsing_error = None
        
        else:
            content_type_parsing_error = ContentTypeParsingError(*error)
        
    return content_type, content_type_parsing_error


class ContentTypeParsingError(RichAttributeErrorBaseType):
    """
    Returned when a content-type could not been fully parsed.
    Contains information to identify error location.
    
    Attributes
    ----------
    expected : `str`
        A string of the expected characters.
    
    index : `int`
        The position where the unexpected character is.
    
    string : `str`
        The string where the error occurred.
    """
    __slots__ = ('expected', 'index', 'string')
    
    def __new__(cls, string, index, expected):
        """
        Creates a new content type parsing error.
        
        Parameters
        ----------
        string : `str`
            The string where the error occurred.
        
        index : `int`
            The position where the unexpected character is.
        
        expected : `str`
            A string of the expected characters.
        """
        self = object.__new__(cls)
        self.expected = expected
        self.index = index
        self.string = string
        return self
    
    
    def __repr__ (self):
        """Returns repr(self)."""
        repr_parts = ['<', type(self).__name__]
        
        # string
        repr_parts.append(' string = ')
        repr_parts.append(repr(self.string))
        
        # index
        repr_parts.append(', index = ')
        repr_parts.append(repr(self.index))
        
        # expected
        repr_parts.append(', expected = ')
        repr_parts.append(repr(self.expected))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns self == other."""
        if type(self) is not type(other):
            return NotImplemented
        
        # string
        if self.string != other.string:
            return False
        
        # index
        if self.index != other.index:
            return False
        
        # expected
        if self.expected != other.expected:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns hash(self)."""
        hash_value = 0
        
        # string
        hash_value ^= hash(self.string)
        
        # index
        hash_value ^= self.index
        
        # expected
        hash_value ^= hash(self.expected)
        
        return hash_value


class ContentType(RichAttributeErrorBaseType):
    """
    A content type, media type, or MIME type represents can be used to identify a data's type and format.
    
    Attributes
    ----------
    parameters : `MultiValueDictionary<str, str>`
        Additional parameters describing the content's format.
    
    sub_type : `None | str`
        Usually consists of the media format.
    
    suffix : `None | str`
        Augmentation for the content to additionally specify its underlying  structure.
    
    type : `None | str`
        The board use of the media type.
    """
    __slots__ = ('parameters', 'sub_type', 'suffix', 'type')
    
    def __new__(cls, type_, sub_type, suffix, parameters):
        """
        Creates a content type with the given parameters.
        
        Parameters
        ----------
        type_ : `None | str`
            The board use of the media type.
        
        sub_type : `None | str`
            Usually consists of the media format.
        
        suffix : `None | str`
            Augmentation for the content to additionally specify its underlying  structure.
        
        parameters : `MultiValueDictionary<str, str>`
            Additional parameters describing the content's format.
        """
        self = object.__new__(cls)
        self.parameters = parameters
        self.sub_type = sub_type
        self.suffix = suffix
        self.type = type_
        return self
    
    
    @classmethod
    def create_empty(cls):
        """
        Creates an empty content type.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.parameters = None
        self.sub_type = None
        self.suffix = None
        self.type = None
        return self
    
    
    def __repr__(self):
        """Returns repr(self)."""
        repr_parts = ['<', type(self).__name__]
        
        # type_
        type_ = self.type
        if (type_ is None):
            field_added = False
        
        else:
            field_added = True
            
            repr_parts.append(' type = ')
            repr_parts.append(repr(type_))
        
        # sub_type
        sub_type = self.sub_type
        if (sub_type is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' sub_type = ')
            repr_parts.append(repr(sub_type))
        
        # suffix
        suffix = self.suffix
        if (suffix is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' suffix = ')
            repr_parts.append(repr(suffix))
        
        # parameters
        parameters = self.parameters
        if (parameters is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' parameters = ')
            repr_parts.append(repr(parameters))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
        
    
    def __eq__(self, other):
        """Returns self == other."""
        if type(self) is not type(other):
            return NotImplemented
        
        # type
        if self.type != other.type:
            return False
        
        # sub_type
        if self.sub_type != other.sub_type:
            return False
        
        # suffix
        if self.suffix != other.suffix:
            return False
        
        # parameters
        if self.parameters != other.parameters:
            return False
        
        return True
    
    
    def __hash__(self, other):
        """Returns hash(self)."""
        hash_value = 0
        
        # type
        type_ = self.type
        if (type_ is not None):
            hash_value ^= hash(type_)
        
        # sub_type
        sub_type = self.sub_type
        if (sub_type is not None) and (sub_type != '*'):
            hash_value ^= hash(sub_type)
        
        # suffix
        suffix = self.suffix
        if (suffix is not None):
            hash_value ^= hash(suffix)
        
        # parameters
        parameters = self.parameters
        if (parameters is not None):
            for key in sorted(parameters.keys()):
                value_hash = 0
                for value in parameters.get_all(key):
                    value_hash ^= hash(value)
                
                hash_value ^= (hash(key) & value_hash)
        
        return hash_value
    
    
    def get_parameter(self, key, default = None):
        """
        Returns the parameter of the content type for the given key.
        
        Parameters
        ----------
        key : `str`
            Key to return the value for.
        
        default : `object` = `None`
            Default value to return if value is missing.
        
        Returns
        -------
        value : `str | default`
        """
        parameters = self.parameters
        if parameters is None:
            return default
        
        return parameters.get(key, default)



from .web_socket import *
from .web_socket import __all__

from warnings import warn

warn(
    f'`websocket` has been renamed to `web_socket`. Only name is deprecated and will be removed in 2025 August.',
    FutureWarning,
    stacklevel = 2,
)



__all__ = ()

from ..utils import RichAttributeErrorBaseType


class ConnectionKey(RichAttributeErrorBaseType):
    """
    Contains information about a host, like proxy, TLS to prevent reusing wrong connections from the pool.
    
    Attributes
    ----------
    host : `str`
        The host's ip address.
    
    port : `int`
        The host's port.
    
    proxy : `None | Proxy`
        Proxy if applicable.
    
    secure : `bool`
        Whether the connection is secure.
    
    ssl_context : `None | SSLContext``
        The connection's ssl type.
    
    ssl_fingerprint : `None | SSLFingerprint`
        Alternative way to accept ssl or to block it depending whether the fingerprint is the same or changed.
    """
    __slots__ = (
        'host', 'port', 'proxy', 'secure', 'ssl_context', 'ssl_fingerprint'
    )
    
    def __new__(cls, host, port, proxy, secure, ssl_context, ssl_fingerprint):
        """
        Creates a new connection key.
        
        Parameters
        ----------
        host : `str`
            The host's ip address.
        
        port : `int`
            The host's port.
        
        proxy : `None | Proxy`
            Proxy if applicable.
        
        secure : `bool`
            Whether the connection is secure.
        
        ssl_context : `None | SSLContext``
            The connection's ssl type.
        
        ssl_fingerprint : `None | SSLFingerprint`
            Alternative way to accept ssl or to block it depending whether the fingerprint is the same or changed.
        """
        self = object.__new__(cls)
        self.host = host
        self.port = port
        self.proxy = proxy
        self.secure = secure
        self.ssl_context = ssl_context
        self.ssl_fingerprint = ssl_fingerprint
        return self
    
    
    def __repr__(self):
        """Returns the connection key's representation."""
        repr_parts = ['<', type(self).__name__]
        
        # host
        repr_parts.append(' host = ')
        repr_parts.append(repr(self.host))
        
        # port
        repr_parts.append(', post = ')
        repr_parts.append(repr(self.port))
        
        # proxy
        proxy = self.proxy
        if (proxy is not None):
            repr_parts.append('proxy = ')
            repr_parts.append(repr(proxy))
        
        # secure
        secure = self.secure
        if secure:
            repr_parts.append(', secure = ')
            repr_parts.append(repr(secure))
        
        # ssl_context
        ssl_context = self.ssl_context
        if (ssl_context is not None):
            repr_parts.append(', ssl_context = ')
            repr_parts.append(repr(ssl_context))
        
        # ssl_fingerprint
        ssl_fingerprint = self.ssl_fingerprint
        if (ssl_fingerprint is not None):
            repr_parts.append(', ssl_fingerprint = ')
            repr_parts.append(repr(ssl_fingerprint))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two connection keys are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # host
        if self.host != other.host:
            return False
        
        # port
        if self.port != other.port:
            return False
        
        # proxy
        if self.proxy != other.proxy:
            return False
        
        # secure
        if self.secure != other.secure:
            return False
        
        # ssl_context
        if self.ssl_context != other.ssl_context:
            return False
        
        # ssl_fingerprint
        if self.ssl_fingerprint != other.ssl_fingerprint:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the connection key's hash value."""
        hash_value = 0
        
        # host
        hash_value ^= hash(self.host)
        
        # port
        hash_value ^= self.port << 17
        
        # proxy
        proxy = self.proxy
        if (proxy is not None):
            hash_value ^= hash(proxy)
        
        # secure
        hash_value ^= self.secure
        
        # ssl_context
        ssl_context = self.ssl_context
        if (ssl_context is not None):
            hash_value ^= hash(ssl_context)
        
        # ssl_fingerprint
        ssl_fingerprint = self.ssl_fingerprint
        if (ssl_fingerprint is not None):
            hash_value ^= hash(ssl_fingerprint)
        
        return hash_value
    
    
    def copy_proxyless(self):
        """
        Copies the connection key without its proxy information.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.host = self.host
        new.port = self.port
        new.proxy = None
        new.secure = self.secure
        new.ssl_context = self.ssl_context
        new.ssl_fingerprint = self.ssl_fingerprint
        return new
